open Types
open Smap

exception Arrived

class packagedb (width, height) =
  object
    val packagedb = Hashtbl.create 101
    val packagemap = Array.make_matrix (width + 1) (height + 1) []

    method find id = let _, _, p = Hashtbl.find packagedb id in p
    method known () = Hashtbl.fold (fun id (x, y, p) l ->
      if List.mem p.package_id packagemap.(x).(y) then p::l else l) packagedb []
    method whereis id = let x, y, p = Hashtbl.find packagedb id in x, y
    method number_of x y = List.length packagemap.(x).(y)
    method found x y packs =
      List.iter (fun p -> Hashtbl.replace packagedb p.package_id (x, y, p)) packs;
      packagemap.(x).(y) <- List.map (fun p -> p.package_id) packs
    method iter f = Hashtbl.iter f packagedb
    method pick x y id =
      packagemap.(x).(y) <- List.filter (fun x -> x <> id) packagemap.(x).(y)

    method drop x y id =
      try let _, _, p = Hashtbl.find packagedb id in
      if p.dest_x = x && p.dest_y = y then Hashtbl.remove packagedb id
      else packagemap.(x).(y) <- id::packagemap.(x).(y)
      with
	(* May be disappeared, or not *)
	Not_found -> packagemap.(x).(y) <- id::packagemap.(x).(y)
  end

class robot (id : int) packagedb =
  object
    val mutable x = 0
    val mutable y = 0
    method id = id
    method x = x
    method y = y
    method symbol = '$'
    method apply_action move =
      let unit_action = function
	  N -> y <- y + 1
	| S -> y <- y - 1
	| E -> x <- x + 1
	| W -> x <- x - 1
	| XY (sx, sy) -> x <- sx; y <- sy
	| PID pid -> packagedb#pick x y pid
	| DID pid -> packagedb#drop x y pid in
      List.iter unit_action move
  end

class robotdb (width, height) packagedb =
  object (self)
    val robotdb = Hashtbl.create 101
    method find id = Hashtbl.find robotdb id
    method iter f = Hashtbl.iter f robotdb
    method exists p = Hashtbl.fold
	(fun key data v -> if p key data then true else v) robotdb false
    method add robot = Hashtbl.replace robotdb robot#id robot

    method update l =
      let rec loop alive l =
	match l with
	  [] ->
	    Hashtbl.iter begin fun id robot ->
	      if not (List.mem id alive) then begin
		Hashtbl.remove robotdb id
	      end
	    end robotdb
	| (id, move)::actions' ->
	    let robot = try Hashtbl.find robotdb id with
	      Not_found ->
		let robot = new robot id packagedb in
		Hashtbl.replace robotdb id robot; robot in
	    robot#apply_action move;
	    loop (id::alive) actions' in
      loop [] l;
  end

(***********)

type plan =
    Initial
  | Move

type idea =
    Visit_home
  | Deliver
  | Pick

let string_of_plan = function
    Initial -> "initial"
  | Move -> "moving"

let string_of_idea = function
    Visit_home -> "visiting home"
  | Deliver -> "delivering"
  | Pick -> "picking"
  

class myrobot (id, max_weight, money) map robotdb packagedb =
  object (self)
    inherit robot id packagedb
    method symbol = '&'
    val mutable weight = 0
    val mutable packages = []
    val mutable money : int = money

	(* Current plan and routing *)
    val mutable plan = Initial
    val mutable route = []
    val mutable dx = 0
    val mutable dy = 0

	(* knowledges about the world *)
    val mutable homes_to_visit = map#home

    method apply_action move =
      let unit_action = function
	  N -> y <- y + 1
	| S -> y <- y - 1
	| E -> x <- x + 1
	| W -> x <- x - 1
	| PID id ->
	    packagedb#pick x y id;
	    let p = packagedb#find id in
	    packages <- p::packages;
	    weight <- weight + p.weight
	| DID id ->
	    Global.debug "dropped";
	    packagedb#drop x y id;
	    packages <- List.filter (fun x ->
	      if x.package_id = id then begin
		weight <- weight - x.weight;
		false
	      end else true) packages
	| XY (sx, sy) -> x <- sx; y <- sy in
      List.iter unit_action move
	
    method set_route dest_x dest_y =
      dx <- dest_x; dy <- dest_y;
      route <- fst (map#dijkstra x y dx dy);

    method follow_the_route () =
      match route with
	[] ->
	  failwith "Empty route";
      | [s] ->
	  if not (Smap.is_in s x y) then begin
	    (* recalculate the route *)
	    route <- fst (map#dijkstra x y dx dy);
	    self#follow_the_route ()
	  end else
	    if dx < x then "Move W" else
	    if x < dx then "Move E" else
	    raise Arrived
      | s::(s'::_ as route') ->
	  if not (Smap.is_in s x y) then begin
	    (* recalculate the route *)
	    route <- fst (map#dijkstra x y dx dy);
	    self#follow_the_route ()
	  end else begin
	    let goup () =
	      route <- route';
              if s'.Smap.y < y then "Move S" else "Move N" in
	    match Smap.margin s' x with
	      Safe -> goup ()
	    | Left_margin ->
		begin match Smap.margin s (x + 1) with
		  Right_margin | Right -> goup ()
		| _ -> "Move E"
		end
	    | Right_margin ->
		begin match Smap.margin s (x - 1) with
		  Left_margin | Left -> goup ()
		| _ -> "Move W"
		end
	    | Right -> "Move W"
	    | Left -> "Move E"
	  end

    method debug npacks =
      Global.debug (Printf.sprintf "money=%d, weight=%d / %d, destination=%d, %d packages=%d found=%d plan=%s"
		      money weight max_weight dx dy (List.length packages) npacks (string_of_plan plan));

    val history = Array.create 6 (0, 0)
    val mutable escape_time = 0

    method think (packs : package list) =
      money <- money - 1;
      self#debug (List.length packs);
      packagedb#found x y packs;
      Array.blit history 0 history 1 (Array.length history - 1);
      history.(0) <- (x, y);
      let rec turn () =
	match plan with
	  Initial -> begin
	    let ideas = List.concat [
	      if weight * 100 < max_weight * 30 then
		List.map (fun (hx, hy) ->
		  Visit_home, hx, hy, map#dijkstra x y hx hy) homes_to_visit
	      else [];
	      List.map (fun p ->
		Deliver, p.dest_x, p.dest_y, map#dijkstra x y p.dest_x p.dest_y) packages;
	      List.fold_right (fun p l ->
		let px, py = packagedb#whereis p.package_id in
		if p.weight + weight > max_weight then l
		else (Pick, px, py, map#dijkstra x y px py)::l) (packagedb#known ()) []] in

	    let ideas = List.stable_sort
		(fun (i, dx, dy, d) (i', dx', dy', d') ->
		  if d < d' then -1 else if d > d' then 1 else 0) ideas in

	    let rec decide = function
		[] -> failwith "No plan"
	      | (Visit_home, dx', dy', (_, _))::ideas' when x = dx' && y = dy' ->
		  homes_to_visit <- List.filter (fun p -> p <> (x, y)) homes_to_visit;
		  if homes_to_visit = [] then homes_to_visit <- map#home;
		  decide ideas'

	      | (Deliver, dx', dy', (_, _))::ideas' when x = dx' && y = dy' ->
		  "Drop" ^ List.fold_right (fun p l ->
		    if p.dest_x = x && p.dest_y = y then " " ^ string_of_int p.package_id ^ l
		    else l) packages ""

	      | (Pick, dx', dy', (_, _))::ideas' when x = dx' && y = dy' ->
		  let s, _ = List.fold_right (fun p (l, w) ->
		    if p.weight + w < max_weight then
		      (" " ^ string_of_int p.package_id ^ l, p.weight + w)
		    else
		      (l, w)) packs ("", weight) in
		  "Pick " ^ s

	      | (i, dx', dy', (r, d))::ideas' ->
		  Global.debug (Printf.sprintf "%d %d %d" dx' dy' d);
		  plan <- Move; route <- r; dx <- dx'; dy <- dy'; turn () in
	    decide ideas
	  end

	| Move ->
	    try self#follow_the_route ()
	    with Arrived ->
	      plan <- Initial; turn () in

      if escape_time = 1 then begin
	escape_time <- escape_time - 1;
	plan <- Initial; "1 " ^ turn ()
      end else begin
	escape_time <- escape_time - 1;
	let c = Array.fold_right (fun (hx, hy) c ->
	  if x = hx && y = hy then c + 1 else c) history 0 in
	if c >= 3 then begin
	  escape_time <- 10;
	  let rec loop () =
	    dx <- Random.int map#width;
	    dy <- Random.int map#height;
	    try let r, d = map#dijkstra x y dx dy in
	    route <- r; plan <- Move; "20 " ^ turn ()
	    with
	      _ -> loop () in
	  loop ()
	end else let is_around rid r = rid <> id && abs (r#x - x) < 4 && abs (r#y - y) < 4 in
	if robotdb#exists is_around then begin
	  Global.debug "rerouting";
	(* Rerouting to avoid collision with enemies *)
	  let newmapstring = String.copy map#mapstring in
	  robotdb#iter begin fun rid r ->
	    if is_around rid r then String.set newmapstring (r#y * (map#width + 2) + r#x) '#';
	  end;
	  let newmap = new Smap.map (map#width, map#height) newmapstring in
	  try
	    route <- fst (newmap#dijkstra x y dx dy);
	    "20 " ^ turn ()
	  with _ ->
	    plan <- Initial; "20 " ^ turn ()
	end else
	  "1 " ^ turn ()
      end
  end
